(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/drivers/mail.js                                              //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
var _this = this;                                                      //
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Crypto = Npm.require("crypto");                                    // 17
var Future = Npm.require("fibers/future");                             // 18
var Promise = Npm.require("es6-promise").Promise;                      // 19
var Capnp = Npm.require("capnp");                                      // 20
                                                                       //
var EmailRpc = Capnp.importSystem("sandstorm/email.capnp");            // 22
var EmailImpl = Capnp.importSystem("sandstorm/email-impl.capnp");      // 23
var HackSessionContext = Capnp.importSystem("sandstorm/hack-session.capnp").HackSessionContext;
var SupervisorCapnp = Capnp.importSystem("sandstorm/supervisor.capnp");
var Supervisor = SupervisorCapnp.Supervisor;                           // 26
var SystemPersistent = SupervisorCapnp.SystemPersistent;               // 27
var EmailSendPort = EmailRpc.EmailSendPort;                            // 28
                                                                       //
var Url = Npm.require("url");                                          // 30
                                                                       //
var ROOT_URL = Url.parse(process.env.ROOT_URL);                        // 32
var HOSTNAME = ROOT_URL.hostname;                                      // 33
                                                                       //
var DAILY_LIMIT = 50;                                                  // 35
var RECIPIENT_LIMIT = 20;                                              // 36
                                                                       //
var CLIENT_TIMEOUT = 15000; // 15s                                     // 38
                                                                       //
// Every day, reset all per-user sent counts to zero.                  //
// TODO(cleanup): Consider a more granular approach. For example, each user could have a timer
//   after which their count will reset. We'd only check the timer when that user is trying to
//   send a new message. This avoids a global query.                   //
if (!Meteor.settings.replicaNumber) {                                  // 44
  // only first replica                                                //
  SandstormDb.periodicCleanup(86400000, function () {                  // 45
    Meteor.users.update({ dailySentMailCount: { $exists: true } }, { $unset: { dailySentMailCount: "" } }, { multi: true });
  });                                                                  //
}                                                                      //
                                                                       //
Meteor.startup(function () {                                           // 52
  var SANDSTORM_SMTP_PORT = parseInt(process.env.SANDSTORM_SMTP_PORT, 10) || 30025;
  var BIND_IP = process.env.BIND_IP || "127.0.0.1";                    // 54
                                                                       //
  simplesmtp.createSimpleServer({ SMTPBanner: "Sandstorm Mail Server" }, function (req) {
    var mailparser = new MailParser();                                 // 57
    req.pipe(mailparser);                                              // 58
                                                                       //
    mailparser.on("end", function (mail) {                             // 60
      // Wrap in outer promise for easier error handling.              //
      Promise.resolve().then(function () {                             // 62
        // Extract the 'from' address.                                 //
        var from = undefined;                                          // 64
        if (mail.from && mail.from.length > 0) {                       // 65
          // It's theoretically possible for the message to have multiple 'from' headers, but this
          // never really happens in legitimate practice so we'll just take the first one.
          from = mail.from[0];                                         // 68
        } else {                                                       //
          // The mail body is missing a 'From:' header. We'll use req.from instead. Note that
          // req.from is actually the bounce address, which is sometimes *not* the original sender
          // but rather some intermediate agent (e.g. a mailing list daemon). See:
          //   http://en.wikipedia.org/wiki/Bounce_address             //
          // TODO(someday): Is this really right, or should we report a blank address instead?
          from = { address: req.from, name: false };                   // 75
        }                                                              //
                                                                       //
        var attachments = [];                                          // 78
        if (mail.attachments) {                                        // 79
          attachments = mail.attachments.map(function (attachment) {   // 80
            var disposition = attachment.contentDisposition || "attachment";
            disposition += ';\n\tfilename="' + (attachment.fileName || attachment.generatedFileName) + '"';
            return {                                                   // 83
              contentType: attachment.contentType,                     // 84
              contentDisposition: disposition,                         // 85
              contentId: attachment.contentId,                         // 86
              content: attachment.content                              // 87
            };                                                         //
          });                                                          //
        }                                                              //
                                                                       //
        if (mail.replyTo && mail.replyTo.length > 1) {                 // 92
          console.error("More than one reply-to address address was received in an email.");
        }                                                              //
                                                                       //
        var mailMessage = {                                            // 96
          // Note that converting the date to nanoseconds actually goes outside the range of
          // integers that Javascript can represent precisely. But this is OK because dates aren't
          // precise anyway.                                           //
          date: (mail.date || new Date()).getTime() * 1000000,         // 100
          from: from,                                                  // 101
          to: mail.to,                                                 // 102
          cc: mail.cc || [],                                           // 103
          bcc: mail.bcc || [],                                         // 104
          replyTo: mail.replyTo && mail.replyTo[0] || {},              // 105
          messageId: mail.headers["message-id"] || Meteor.uuid() + "@" + HOSTNAME,
          references: mail.references || [],                           // 107
          inReplyTo: mail.inReplyTo || [],                             // 108
          subject: mail.subject || "",                                 // 109
          text: mail.text || null,                                     // 110
          html: mail.html || null,                                     // 111
          attachments: attachments                                     // 112
        };                                                             //
                                                                       //
        // Get list of grain IDs.                                      //
        var grainPublicIds = _.uniq(req.to.map(function (deliverTo) {  // 116
          // simplesmtp already validates that the address contains an @.
          // To simplify things, we ignore the hostname part of the address and assume that the
          // message would not have been sent here if it weren't intended for our host. Usually
          // there will be an nginx frontend verifying hostnames anyway. Grain public IDs are
          // globally unique anyway, so an e-mail meant for another server presumably won't match
          // any ID at this one anyway.                                //
          return deliverTo.slice(0, deliverTo.indexOf("@"));           // 123
        }));                                                           //
                                                                       //
        // Deliver to each grain in parallel.                          //
        return Promise.all(grainPublicIds.map(function (publicId) {    // 127
          // Wrap in a function so that we can call it recursively to retry.
          var tryDeliver = function (retryCount) {                     // 129
            var grainId = undefined;                                   // 130
            return inMeteor(function () {                              // 131
              var grain = Grains.findOne({ publicId: publicId }, { fields: {} });
              if (grain) {                                             // 133
                grainId = grain._id;                                   // 134
                return globalBackend.continueGrain(grainId, retryCount > 0);
              } else {                                                 //
                // TODO(someday): We really ought to rig things up so that the 'RCPT TO' SMTP command
                //   fails in this case, but simplesmtp doesn't appear to support that.
                throw new Error("No such grain: " + publicId);         // 139
              }                                                        //
            }).then(function (grainInfo) {                             //
              var supervisor = grainInfo.supervisor;                   // 142
              var uiView = supervisor.getMainView().view;              // 143
                                                                       //
              // Create an arbitrary struct to use as the session params. E-mail sessions actually
              // require no params, but node-capnp won't let us pass null and we don't have an
              // EmptyStruct type available, so we just use EmailAddress, but any struct type would
              // work.                                                 //
              // TODO(cleanup): Fix node-capnp to accept null.         //
              var emptyParams = Capnp.serialize(EmailRpc.EmailAddress, {});
                                                                       //
              // Create a new session of type HackEmailSession. This is a short-term hack until
              // persistent capabilities and the Powerbox are implemented. We need to pass along a
              // HackSessionContext because old versions of sandstorm-http-bridge always use
              // the context of the most recent session; if an old RoundCube grain has a
              // WebSession open, receives an email, and then tries to send an email, the request
              // will go out on the SessionContext associated with the HackEmailSession that
              // delivered the email.                                  //
              var session = uiView.newSession({}, makeHackSessionContext(grainId), "0xc3b5ced7344b04a6", emptyParams).session.castAs(EmailSendPort);
              return session.send(mailMessage);                        // 163
            })["catch"](function (err) {                               //
              if (SandstormBackend.shouldRestartGrain(err, retryCount)) {
                return tryDeliver(retryCount + 1);                     // 166
              } else {                                                 //
                throw err;                                             // 168
              }                                                        //
            });                                                        //
          };                                                           //
                                                                       //
          return tryDeliver(0);                                        // 173
        }));                                                           //
      }).then(function () {                                            //
        req.accept();                                                  // 176
      }, function (err) {                                              //
        console.error("E-mail delivery failure:", err.stack);          // 178
        req.reject(err.message);                                       // 179
      });                                                              //
    });                                                                //
  }).listen(SANDSTORM_SMTP_PORT, BIND_IP);                             //
});                                                                    //
                                                                       //
function formatAddress(field) {                                        // 185
  if (!field) {                                                        // 186
    return null;                                                       // 187
  }                                                                    //
                                                                       //
  if (Array.isArray(field)) {                                          // 190
    return field.map(formatAddress);                                   // 191
  }                                                                    //
                                                                       //
  if (field.name) {                                                    // 194
    return field.name + " <" + field.address + ">";                    // 195
  }                                                                    //
                                                                       //
  return field.address;                                                // 198
}                                                                      //
                                                                       //
hackSendEmail = function (session, email) {                            // 201
  return inMeteor((function () {                                       // 202
    var recipientCount = 0;                                            // 203
    recipientCount += email.to ? email.to.length : 0;                  // 204
    recipientCount += email.cc ? email.cc.length : 0;                  // 205
    recipientCount += email.bcc ? email.bcc.length : 0;                // 206
    if (recipientCount > RECIPIENT_LIMIT) {                            // 207
      throw new Error("Sorry, Sandstorm currently only allows you to send an e-mail to " + RECIPIENT_LIMIT + " recipients at a time, for spam control. Consider setting up a mailing list. " + "Please feel free to contact us if this is a problem for you.");
    }                                                                  //
                                                                       //
    // Overwrite the 'from' address with the grain's address.          //
    if (!email.from) {                                                 // 215
      email.from = {};                                                 // 216
    }                                                                  //
                                                                       //
    var grainAddress = session._getAddress();                          // 219
    var userAddress = session._getUserAddress();                       // 220
                                                                       //
    // First check if we're changing the from address, and if so, move it to reply-to
    if (email.from.address !== grainAddress && email.from.address !== userAddress.address) {
      throw new Error("FROM header in outgoing emails need to equal either " + grainAddress + " or " + userAddress.address + ". Yours was: " + email.from.address);
    }                                                                  //
                                                                       //
    var mc = new MailComposer();                                       // 229
                                                                       //
    mc.setMessageOption({                                              // 231
      from: formatAddress(email.from),                                 // 232
      to: formatAddress(email.to),                                     // 233
      cc: formatAddress(email.cc),                                     // 234
      bcc: formatAddress(email.bcc),                                   // 235
      replyTo: formatAddress(email.replyTo),                           // 236
      subject: email.subject,                                          // 237
      text: email.text,                                                // 238
      html: email.html                                                 // 239
    });                                                                //
                                                                       //
    var envelope = mc.getEnvelope();                                   // 242
    envelope.from = grainAddress;                                      // 243
                                                                       //
    mc.setMessageOption({                                              // 245
      envelope: envelope                                               // 246
    });                                                                //
                                                                       //
    var headers = {};                                                  // 249
    if (email.messageId) {                                             // 250
      mc.addHeader("message-id", email.messageId);                     // 251
    }                                                                  //
                                                                       //
    if (email.references) {                                            // 254
      mc.addHeader("references", email.references);                    // 255
    }                                                                  //
                                                                       //
    if (email.messageId) {                                             // 258
      mc.addHeader("in-reply-to", email.inReplyTo);                    // 259
    }                                                                  //
                                                                       //
    if (email.date) {                                                  // 262
      var date = new Date(email.date / 1000000);                       // 263
      if (!isNaN(date.getTime())) {                                    // 264
        // Check to make sure date is valid                            //
        mc.addHeader("date", date.toUTCString());                      // 265
      }                                                                //
    }                                                                  //
                                                                       //
    if (email.attachments) {                                           // 269
      email.attachments.forEach(function (attachment) {                // 270
        mc.addAttachment({                                             // 271
          cid: attachment.contentId,                                   // 272
          contentType: attachment.contentType,                         // 273
          contentDisposition: attachment.contentDisposition,           // 274
          contents: attachment.content                                 // 275
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    var grain = Grains.findOne(session.grainId);                       // 280
    if (!grain) throw new Error("Grain does not exist.");              // 281
                                                                       //
    var user = Meteor.users.findAndModify({                            // 283
      query: { _id: grain.userId },                                    // 284
      update: {                                                        // 285
        $inc: {                                                        // 286
          dailySentMailCount: 1                                        // 287
        }                                                              //
      },                                                               //
      fields: { dailySentMailCount: 1 }                                // 290
    });                                                                //
    if (user.dailySentMailCount >= DAILY_LIMIT) {                      // 292
      throw new Error("Sorry, you've reached your e-mail sending limit for today. Currently, Sandstorm " + "limits each user to " + DAILY_LIMIT + " e-mails per day for spam control reasons. " + "Please feel free to contact us if this is a problem.");
    }                                                                  //
                                                                       //
    SandstormEmail.rawSend(mc);                                        // 299
  }).bind(_this))["catch"](function (err) {                            //
    console.error("Error sending e-mail:", err.stack);                 // 301
    throw err;                                                         // 302
  });                                                                  //
};                                                                     //
                                                                       //
var EmailVerifierImpl = (function () {                                 //
  function EmailVerifierImpl(persistentMethods, id, params) {          // 307
    babelHelpers.classCallCheck(this, EmailVerifierImpl);              //
                                                                       //
    _.extend(this, persistentMethods);                                 // 308
    this._id = id;                                                     // 309
    this._services = params.services;                                  // 310
  }                                                                    //
                                                                       //
  EmailVerifierImpl.prototype.getId = (function () {                   // 306
    function getId() {                                                 // 313
      return { id: new Buffer(this._id, "base64") };                   // 314
    }                                                                  //
                                                                       //
    return getId;                                                      //
  })();                                                                //
                                                                       //
  EmailVerifierImpl.prototype.verifyEmail = (function () {             // 306
    function verifyEmail(tabId, verification) {                        // 317
      var _this2 = this;                                               //
                                                                       //
      // For now, we save() the verification and then dig through ApiTokens to find where it leads.
      // TODO(cleanup): In theory we should be using something like CapabilityServerSet, but it is
      //   not available in Javascript yet and even if it were, it wouldn't work in the case where
      //   there are multiple front-end replicas, since the verification could be on a different
      //   replica.                                                    //
      return verification.castAs(SystemPersistent).save({ frontend: null }).then(function (saveResult) {
        return inMeteor(function () {                                  // 324
          var tokenId = hashSturdyRef(saveResult.sturdyRef);           // 325
          var tokenInfo = ApiTokens.findOne(tokenId);                  // 326
                                                                       //
          // Delete the token now since it's not needed.               //
          ApiTokens.remove(tokenId);                                   // 329
                                                                       //
          for (;;) {                                                   // 331
            if (!tokenInfo) throw new Error("missing token?");         // 332
            if (!tokenInfo.parentToken) break;                         // 333
            tokenInfo = ApiTokens.findOne(tokenInfo.parentToken);      // 334
          }                                                            //
                                                                       //
          if (!tokenInfo.frontendRef || !tokenInfo.frontendRef.verifiedEmail) {
            throw new Error("not a VerifiedEmail capability");         // 338
          }                                                            //
                                                                       //
          var verification = tokenInfo.frontendRef.verifiedEmail;      // 341
          if (verification.tabId !== tabId.toString("hex")) {          // 342
            throw new Error("VerifiedEmail is from a different tab");  // 343
          }                                                            //
                                                                       //
          if (_this2._services) {                                      // 346
            // Since this verifier is restricted to specific services, only indicate a match if the
            // VerifiedEmail was for the correct verifier ID. (If our _services is null, then we
            // match all services, and therefore all VerifiedEmails.)  //
            if (verification.verifierId !== _this2._id) {              // 350
              throw new Error("VerifierEmail is for a different EmailVerifier.");
            }                                                          //
          }                                                            //
                                                                       //
          return verification.address;                                 // 355
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return verifyEmail;                                                //
  })();                                                                //
                                                                       //
  return EmailVerifierImpl;                                            //
})();                                                                  //
                                                                       //
;                                                                      // 359
                                                                       //
var VerifiedEmailImpl = (function () {                                 //
  function VerifiedEmailImpl(persistentMethods, id) {                  // 362
    babelHelpers.classCallCheck(this, VerifiedEmailImpl);              //
                                                                       //
    _.extend(this, persistentMethods);                                 // 363
  }                                                                    //
                                                                       //
  return VerifiedEmailImpl;                                            //
})();                                                                  //
                                                                       //
makeEmailVerifier = function (persistentMethods, id, params) {         // 367
  return new Capnp.Capability(new EmailVerifierImpl(persistentMethods, id, params), EmailImpl.PersistentEmailVerifier);
};                                                                     //
                                                                       //
makeVerifiedEmail = function (persistentMethods) {                     // 372
  return new Capnp.Capability(new VerifiedEmailImpl(persistentMethods), EmailImpl.PersistentVerifiedEmail);
};                                                                     //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=mail.js.map
